#include <signal.h>
#include <curses.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>

/*---------------------------------
       K O N S T A N T Y
-----------------------------------*/
#define CHYBA        (-1)       /* chybicka */
#define MOD          0666       /* prava vytvarene fronty zprav */
#define MAX_MSG_LEN   20        /* maximalni delka posilane zpravicky */

#define MS_WAIT       0         /* mod ziskavani zpravy, kdy se ceka az zprava daneho typu prijde */
#define MS_NOWAIT     1         /*  -- || --- neceka se, vrati se chyba */

#define MSG_DISCONNECT "Protihrac se odpojil\n"   /* zprava pri odpojeni druheho hrace

/* Typy zpravicek:  */
#define T_INITAPP     10  /* inicialisace programu (navazani spojeni) */
#define T_INITGAME    3   /* inicialisace nove hry */
#define T_TAH         2   /* normalni tah */
#define T_ENDAPP      1   /* ukonceni porgramu */
#define T_NOTREGISTRED 20

/* souradnice nekterych aktualizovanych udaju */
#define N_X     69
#define N_WIN   13
#define N_LOSS  14
#define N_NTAH  15
#define N_LTAHX 16
#define N_LTAHY 17

/* souradnice a rozmery menu */
#define MENU_X   48
#define MENU_Y   2
#define MENU_W   27

/* sour. a rozmery herniho pole */
#define FIELD_X  4
#define FIELD_Y  2
#define FIELD_W  20
#define FIELD_H  20
#define FIELD_VLINE ':'
#define FIELD_HLINE '_'
#define PISQ         5
#define S_CROSS     'x'
#define S_CIRCLE    'o'
#define S_FREE       FIELD_HLINE

/*-----------------------------------------*/

/*----- Struktura pro zpravicku ------*/
typedef struct {
  long type;
  char text[MAX_MSG_LEN];
} tmsg;

/*----- Globalni promenne ------------*/
int my_sign, his_sign;           /* Moje hraci znacka, protihracova znacka */
char field [FIELD_H][FIELD_W];   /* Pole obsahujici vsechny tahy */
key_t r_key, w_key;              /* Klic fronty pro cteni a pro zapis*/
int server;                      /* Jestli to bude server nebo klient */
int r_queue=0, w_queue=0;        /* Identifikatory front zprav pro zapis a cteni */
int wins, losses;                /* Celkovy pocet proher (vyher) */
int last_x, last_y;              /* Souradnice posledniho tahu */
int tah;                         /* Cislo aktualniho tahu */       
int WinActive=0;                 /* Jestli uz je zaktivovany graficky rezim */    

/*=========================================*/

int newgame_start ();
int get_msg (int type, char *message, int nowait);
void app_end ();

int read_key (){
  int Key = ERR;
  while ((Key = getch ())==ERR);
  return (Key);
}

static void finish (int sig){
  app_end ("Nasilne ukonceni programu Pisqorky\n");
}

/*-----------------------------------------
  Vypise napovedu (help)
------------------------------------------*/
void help (){
  printf ("   .------------------------------.\n");
  printf ("   |      PISQORKY verse 1.0      |\n");
  printf ("   `------------------------------'\n");
  printf ("     Petr Vaclavek\n");
  printf ("     bla@atrey.karlin.mff.cuni.cz\n");
  printf ("     atrey.karlin.mff.cuni.cz/~bla\n\n");
  printf ("  Parametry:\n");
  printf ("     pisqorky -c/-s n\n");
  printf ("                \\/  `cislo kanalu (standartne 0)\n");
  printf ("                 ` klient / server (jeden z hracu \n");
  printf ("                   musi byt klient, druhy server)\n");
  printf ("  Blizsi informace v dokumentaci.\n\n");    
}

/*-----------------
   yx_to_mess (zakoduje souradnice do zpravicky)
-------------------*/
char *yx_to_mess (int y, int x, char *mess) {
  sprintf (mess, "%d|%d", y, x);
  return (mess);
} 
       
/*-----------------
   mess_to_yx   (rozkoduje zpravu do souradnic)
-------------------*/
void mess_to_yx (int *y, int *x, char *mess) {
  sscanf (mess, "%d|%d", y, x);
} 
           

/*-----------------------------------------
   Vycisteni pole tahu (Hraci plochy) 
-------------------------------------------*/
void clear_field (){
  int i,j;
  for (i=0; i<FIELD_H; i++)
    for (j=0; j<FIELD_W; j++)
      field [i][j] = S_FREE;
}      
   
/*-------------------------------
    Uvolneni front zprav
---------------------------------*/
void free_qs (int queue_id){
  if (msgctl (queue_id, IPC_RMID, 0) == CHYBA){
    printf ("Chyba pri ruseni fronty c.%d\n", queue_id);
    exit (1);
  }
}   

/*--------------------------------
  Ukonceni programu
---------------------------------*/
void app_end (char *Mess){
  if (WinActive) {
    clear ();
    refresh ();
    endwin ();
  }  
  if (server){  /* Zrusi fronty zprav */
    if (r_queue) free_qs (r_queue);
    if (w_queue) free_qs (w_queue);
  }
  else {  /* Posle zpravu, ze uz skoncil */
    if (w_queue) send_msg (T_ENDAPP, ""); 
  }

  if (Mess != NULL)
    printf (Mess);
  if (wins || losses)
    printf ("Celkove skore: %d : %d\n", wins, losses);
  exit (0);
}

/*-----------------------------------------
   Vykresleni hraciho pole
-------------------------------------------*/
void paint_field (){
  int i=0,j=0;
  char Pom[4];
  
  mvaddch (FIELD_Y-1, FIELD_X-1, '.');
  mvaddch (FIELD_Y-1, FIELD_X+2*(FIELD_W)-1, '.');
  move (FIELD_Y-1,FIELD_X);
  hline (FIELD_HLINE,2*FIELD_W-1);
  for (j=0;j<FIELD_H;j++){
    sprintf (Pom, "%d", j%10);
    mvaddstr (FIELD_Y+j, FIELD_X-2,Pom);
    mvaddstr (FIELD_Y-1, FIELD_X+2*j, Pom);
    for (i=0;i<FIELD_W*2;i++)
      (i%2)?mvaddch (j+FIELD_Y,i+FIELD_X,FIELD_VLINE):mvaddch(j+FIELD_Y,i+FIELD_X,FIELD_HLINE);
  }    
      

  move (FIELD_Y,FIELD_X-1);
  vline (FIELD_VLINE, FIELD_H);
    
  refresh ();
}

/*----------------------------
  vypsani nejakeho cisla do prislusneho radku
-----------------------------*/
void print_num (int num, int where, int refr){
  char Pom[10];
  int x, y;
  getyx (stdscr, y, x);
  sprintf (Pom, "%d", num);
  move (where, N_X);
  hline (' ',4);
  mvaddstr (where, N_X, Pom);
  move (y,x);
  if (refr) refresh ();
}

/*------------------------------------------
   Vykresleni uvodni obrazovky
-------------------------------------------*/
void paint_init_screen (){
  clear ();
  border ('|','|','-','-','.','.','`','\'');
  move (MENU_Y, MENU_X);
  hline ('*', MENU_W);
  mvaddstr (MENU_Y+1, MENU_X, "***** P I S Q O R K Y *****");
  move (MENU_Y+2, MENU_X);
  hline ('*', MENU_W);
  move (MENU_Y+3, MENU_X);
  mvaddstr (MENU_Y+3, MENU_X, "verse 1.0  (26. ledna 1998)");
  mvaddstr (MENU_Y+4, MENU_X, "       Petr Vaclavek       ");
  mvaddstr (MENU_Y+6, MENU_X, "Ovladani:");
  mvaddstr (MENU_Y+7, MENU_X, "q - konec hry");
  mvaddstr (MENU_Y+8, MENU_X, "Sipky - pohyb");
  mvaddstr (MENU_Y+9, MENU_X, "Mezernik - polozeni znacky");
  move (MENU_Y+10, MENU_X);
  hline ('-', MENU_W);
  mvaddstr (N_WIN,   MENU_X, "   Pocet vyher     :"); 
  mvaddstr (N_LOSS,  MENU_X, "   Pocet proher    :");
  mvaddstr (N_NTAH,  MENU_X, "   Pocet tahu      :"); 
  mvaddstr (N_LTAHX, MENU_X, "   Posledni tah (X):"); 
  mvaddstr (N_LTAHY, MENU_X, "   Posledni tah (Y):"); 
  move (MENU_Y+16, MENU_X);
  hline ('-', MENU_W);
  print_num (wins, N_WIN, 0);
  print_num (losses, N_LOSS, 0);
  print_num (tah, N_NTAH, 0);
  print_num (last_x, N_LTAHX, 0);
  print_num (last_y, N_LTAHY, 0);
  refresh ();
}

/*-------------------------------------
  Vypsani hlasky
---------------------------------------*/
void print_message (char *message, int Line){
  int Len, X;
  int x,y;
  getyx (stdscr, y, x);
  Len=strlen (message);
  if (Len > MENU_W)
    Len = MENU_W;
    
  X = MENU_X + (MENU_W - Len)/2;  
  move (MENU_Y + Line + 18, MENU_X);
  hline (' ', MENU_W);
  mvaddnstr (MENU_Y + Line + 18, X, message, Len);
  move (y,x);
  refresh ();
}

  
/*-----------------------------------------
  Zjisti a prevede pozici kursoru na souradnice v hracim poli
------------------------------------------*/
void get_fyfx (int *fy, int *fx){
  int x,y;
  getyx (stdscr, y, x);
  y-=FIELD_Y;
  x-=FIELD_X;
  x=x/2;
  *fx=x;
  *fy=y;
}

/*---------------------------------------
  Presune kursor na hracim poli na souradnice fy, fx
-----------------------------------------*/
void  move_to_field (int fy, int fx){
  int x,y;
  x=FIELD_X+2*fx;
  y=FIELD_Y+fy;
  move (y,x);
/*  refresh (); */
}

/*--------------------------------
  Polozi na souradnice fy, fx znacku "s"
  (Zkontroluje, zda je tam volno, pokud ne tak pipne a nic neudela)
----------------------------------*/  
void put_sign (int fy, int fx, int s){
  move_to_field (fy, fx);
  addch (s);
  field [fy][fx]=s;
  move_to_field (fy, fx);
}

int get_sign (int fy, int fx){
  return (field [fy][fx]);
}

/*--------------------------
   Poslani tahu protivnikovi
----------------------------*/
void send_tah (int y, int x){
  char Pom[20];
  yx_to_mess (y, x, Pom);
  send_msg (T_TAH, Pom);
}

/*-------------------------
  Ziskani protivnikova tahu
---------------------------*/
void get_tah (int *y, int *x){
  char Pom[20];
  print_message ("Protivnik je",0);
  print_message ("na tahu",1);
  get_msg (-T_TAH, Pom, MS_WAIT);
  mess_to_yx (y, x, Pom);
  print_message ("Jsi na tahu", 0);
  print_message ("Hrajes!", 1);
}


/*---------------- IS_GAMEOVER? ----------------
  Zjisti, zda uz jsme dohrali 
--------------------------------------------*/
int is_gameover (int fy, int fx, int sign){
  int Count=0;       // Pocet znaku
  /* vymezeni ctverce, kde se bude kontrolovat, zda uz jsme dohrali */
  int start_x, start_y, stop_x, stop_y; 
  int i=0;
  
  start_x = fx - PISQ;
  start_y = fy - PISQ;
  stop_x  = fx + PISQ;
  stop_y  = fy + PISQ;
  if (start_x < 0) start_x = 0;
  if (start_y < 0) start_y = 0;
  if (stop_x > FIELD_W) stop_x = FIELD_W;
  if (stop_y > FIELD_H) stop_y = FIELD_H;
  
  /* Musim zkontrolovat 4 smery ( / \ | -- )  */
  /* 1) Vodorovne */
  Count = 0;
  for (i = start_x; i <= stop_x; i++){
    if (field [fy][i] == sign)
      Count++;
    else
      Count = 0;
    if (Count >= PISQ) return (1);
  }   

  /* 2) Svisle */
  Count = 0;
  for (i = start_y; i <= stop_y; i++){
    if (field [i][fx] == sign)
      Count++;
    else
      Count = 0;
    if (Count >= PISQ) return (2);
  }   
  
  
  /* Ted to bude krapet komplikovanejsi - sikme smery */
  /* 3) Vpravo dolu.... */
  Count=1;
  for (i=1; (fx-i >= start_x && fy-i >= start_y); i++){
    if (field [fy-i][fx-i] == sign)
      Count++;
    else
      break;  
  }
  for (i=1; (fx+i <= stop_x && fy+i <= stop_y); i++){
    if (field [fy+i][fx+i] == sign)
      Count++;
    else
      break;
  }
  
  if (Count >= PISQ) return (3);
  
  
  /* 4) Vlevo dolu a vpravo nahoru.... */
  Count=1;
  for (i=1; (fx-i >= start_x) && (fy+i <= stop_y); i++){
    if (field [fy+i][fx-i] == sign)
      Count++;
    else
      break;
  }
  for (i=1; (fx+i <= stop_x) && (fy-i >= start_y); i++){
    if (field [fy-i][fx+i] == sign)
      Count++;
    else
      break;
  }
  
  if (Count >= PISQ) return (4);
  else return (0);
}

/*----------------------------
  Game Over - aneb co se provede, pokud nekdo vyhraje
------------------------------*/
void game_over (int sign){
  if (sign == my_sign){
    wins++;
    print_num (wins,N_WIN, 0);
    print_message ("Vyhral jsi!", 0);
  }  
  else {
    losses++;
    print_num (wins,N_WIN, 0);
    print_message ("Prohral jsi!", 0);
  }  
  print_message ("Zmackni cokoliv", 1);
  read_key ();
}


/*-------------------------------
  Provedeni nejakeho tahu
--------------------------------*/
void make_tah (int fy, int fx, int sign){
  put_sign (fy, fx, sign);
  tah++;
  print_num (tah, N_NTAH, 0);
  print_num (fy, N_LTAHY, 0);
  print_num (fx, N_LTAHX, 1);
}

/*-------------------------------
  Ocetreni vstupu z klavesnice
---------------------------------*/
int key_pressed (int Key, int *y, int *x){
  int fx,fy;
  get_fyfx (&fy, &fx);
  
  if (Key==KEY_LEFT)
    if (fx>0) fx--;
    else fx=FIELD_W-1;  
  if (Key==KEY_RIGHT) 
    if (fx<FIELD_W-1) fx++;
    else fx=0;  
  if (Key==KEY_UP)
    if (fy>0) fy--;
    else fy=FIELD_H-1;
  if (Key==KEY_DOWN)
    if (fy<FIELD_H-1) fy++;
    else fy=0;

  
  *x = fx;
  *y = fy;
  if (Key==32){
    if (field [fy][fx] == S_FREE){
      make_tah (fy, fx, my_sign);
      return (1);
    }  
    else {
      beep (); 
      refresh ();
    }  
  }     

  if (Key != 32) move_to_field (fy, fx); 
  return (0);
}

/*------------------------------
    Vytvori fronty zprav pro cteni a zapis
-------------------------------*/
int make_qs (int key){
  /* Zjisti se zda uz nahodou dana fronta neexistuje */
  int Pom=0;
  
  if (server){   /* SERVER */
    if (msgget (key,0)==CHYBA && errno==ENOENT){
      /* fronta neexistujeme, muzeme ji vytvorit */
      Pom=msgget (key, IPC_CREAT | MOD);
      if (Pom == CHYBA) {
        printf ("Server nemuze vytvorit frontu\n");
        app_end (NULL);
      }
    }
    else { /* fronta jiz existuje... */
      printf ("Nekdo uz poziva tuto frontu zprav\n");
      app_end (NULL);
    }  
  }  
  
  else {   /* CLIENT */
    Pom=msgget (key,0);
    if (Pom==CHYBA && errno==ENOENT){
      printf ("Server jeste nevytvoril frontu zprav\n");
      app_end (NULL);
    }  
  }
  return (Pom);
}

/*-------------------------------
  Inicialisace celeho PROGRAMu
---------------------------------*/
void app_start (int argc, char *argv[]){
  char *Pom;
  char *Msg="Start";
  int Chanel = 0;
  
/* Inicialisace roury zprav */
  if (argc<2){
    printf ("CHYBA: Malo parametru!\n\n");
    help ();
    app_end (NULL);
  }

/* Zpracovani parametru */  
  Pom=argv[1];
  if (Pom[1]=='s') server=1;
  else server=0;
  
/* Pripadne cislo kanalu */  
  if (argc>2){
    sscanf (argv[2], "%d", &Chanel);
  }

/* zjisteni  klicu k frontam */
  r_key=ftok (argv[0], 2*Chanel+ ((0+server)%2));
  w_key=ftok (argv[0], 2*Chanel+ ((1+server)%2));
  
/* vytvoreni front */  
  r_queue=make_qs (r_key);
  w_queue=make_qs (w_key);    
  
/* navazani kontaktu */
  if (server){
    send_msg (T_INITAPP, Msg);
    printf ("Cekam na protihrace....\n");
    get_msg (T_INITAPP, Pom, MS_WAIT);
    printf ("Priotihrac pipraven....\n");
  }
  else {  /* client */
    printf ("Pokousim se napojit na protihrace...\n");
    if (get_msg (T_INITAPP, Pom, MS_NOWAIT) ==CHYBA){
      printf ("Tuto frontu zprav jiz nekdo pouziva!\n");
      printf ("Zkuste to jindy, nebo na jinem kanale\n");
      w_queue = 0; r_queue = 0;
      app_end (NULL); 
    }
    printf ("Protihrac pripraven...\n");
    send_msg (T_INITAPP, Msg);
  }  
  
/* inicialisace promennych */  
  wins = 0;
  losses = 0;
  if (server){
    my_sign = S_CROSS; 
    his_sign = S_CIRCLE;
  }
  else {
    my_sign = S_CIRCLE;
    his_sign = S_CROSS;
  }

  initscr ();
  WinActive = 1;
  nodelay (stdscr, 1);
  keypad (stdscr, TRUE);
  cbreak ();
  noecho ();

}


/*------------------------------
  Inicialisace NOVE HRY - pokud vraci 1, zacina (jinak zacina protihrac)
-------------------------------*/
int newgame_start (){
/*  int MsgType; */
  char Pom[20];
  char YouBegin[]="Zacinas";
  char HeBegin []="Zacina protihrac";
  tah = 0;
  last_x = 0;
  last_y = 0;
  
  clear_field ();
  paint_init_screen ();
  paint_field ();  
  move_to_field (0,0);  
  print_message ("Zacatek nove hry",0);
  
/* Rozhodnuti, kdo bude zacinat */
  if (server) {
    if (rand() % 2){
      /* zacina server */
      send_msg (T_INITGAME, HeBegin);
      print_message (YouBegin,1);
      return (1);
    }
    else {
      /* zacina klient */
      send_msg (T_INITGAME, YouBegin);
      print_message (HeBegin, 1);
      return (0);
    }
  }
  else { /* klient */
    print_message ("Ceka se na protihrace", 1);
/*    MsgType = -T_INITGAME; */
    get_msg (-T_INITGAME, Pom ,MS_WAIT);
    print_message (Pom, 1);
    if (!strcmp (Pom, YouBegin)){
      return (1);
    }
    else return (0);
  }
}

/*----------------------------
    Poslani zpravicky
-----------------------------*/
int send_msg (int type, char *message){
  tmsg msg;
  int delka=0;
    
  /* typ zpravicky */
  msg.type = type;
  
  /* text zpravicky */
  strncpy (msg.text, message, MAX_MSG_LEN);
  delka = strlen (msg.text);
  if (delka == 0) delka++;
  
  if (msgsnd (w_queue, (void*)&msg, delka,0) == CHYBA){
    /*perror ("[[Nepodarilo se poslat zpravu]]\n");*/
    if (type != T_ENDAPP)
      app_end (MSG_DISCONNECT);
    return (CHYBA);
  }      
  else return (1);
}

/*---------------------------
    Precteni zpravy
-----------------------------*/
int get_msg (int type, char *message, int nowait){
  int res=0;
  tmsg msg;
  int typ=T_NOTREGISTRED;
  int flag = MSG_NOERROR;
  if (nowait) flag |= IPC_NOWAIT;

  while (abs (typ) != abs (type)){ 
    if (WinActive) {
      curs_set(0);
      refresh ();
    }
    res = msgrcv (r_queue, (void *)&msg, sizeof (msg.text), type, flag);
    if (WinActive) {
      curs_set (1);
      refresh ();
    }  
    if (res>0) {
      msg.text [res]='\0';
      strcpy (message, msg.text);
      typ = msg.type;
      if (msg.type == T_ENDAPP) app_end (MSG_DISCONNECT);
    }  
    else if (!nowait){ /* chyba a nastaveno ze nama cekat -> protihrac se odpojil */
      app_end (MSG_DISCONNECT);    
    }
    else return (res); /* chyba a nowait - test, zda je tam protivnik - protivnik uz s nekym hraje*/
  } /* while */
  return (res);  
}


/*===============================
         M A I N ()
=================================*/         

void main (int argc, char *argv[]){
  int Key=0;
  int MyWin=0;
  char Pom[20];
  int y, x;
  
  (void) signal (SIGINT, finish);
  app_start (argc, argv);
 
  if (!newgame_start ()){
    get_tah (&y, &x);
    make_tah (y, x, his_sign);
  }  

  while ((Key=read_key ())!=113){
    if (key_pressed (Key, &y, &x)){  /* provedl nejakej tah */
      send_tah (y,x);
      MyWin=is_gameover (y, x, my_sign);
      if (!MyWin){
        get_tah (&y, &x);
        make_tah (y, x, his_sign);
      }
      
      if (MyWin || is_gameover (y, x, his_sign)){  /* nekdo vyhral... */
        if (MyWin) game_over (my_sign);
        else game_over (his_sign);
        
        if (!newgame_start ()){
          get_tah (&y, &x);
          make_tah (y, x, his_sign);
        } /* end - init_new_game */  
      } /* end - gameover */              
    } /* end - tah */
  } /* end - while */

  app_end ("Diky za hru... \n");
} /* end - prg */

